/*
 *  TOPPERS Software
 *      Toyohashi Open Platform for Embedded Real-Time Systems
 *
 *  Copyright (C) 2007-2008 by TAKAGI Nobuhisa
 *  Copyright (C) 2010 by Meika Sugimoto
 *
 *  嵭Ԥϡʲ(1)(4)ξ˸¤ꡤܥեȥ
 *  ܥեȥѤΤޤࡥʲƱˤѡʣ
 *  ѡۡʰʲѤȸƤ֡ˤ뤳Ȥ̵ǵ롥
 *  (1) ܥեȥ򥽡ɤηѤˤϡ嵭
 *      ɽѾ浪Ӳ̵ݾڵ꤬Τޤޤηǥ
 *      ˴ޤޤƤ뤳ȡ
 *  (2) ܥեȥ򡤥饤֥ʤɡ¾Υեȥȯ˻
 *      ѤǤǺۤˤϡۤȼɥȡ
 *      ԥޥ˥奢ʤɡˤˡ嵭ɽѾ浪Ӳ
 *      ̵ݾڵǺܤ뤳ȡ
 *  (3) ܥեȥ򡤵Ȥ߹ʤɡ¾Υեȥȯ˻
 *      ѤǤʤǺۤˤϡΤ줫ξ
 *      ȡ
 *    (a) ۤȼɥȡѼԥޥ˥奢ʤɡˤˡ嵭
 *        ɽѾ浪Ӳ̵ݾڵǺܤ뤳ȡ
 *    (b) ۤη֤̤ˡˤäơTOPPERSץȤ
 *        𤹤뤳ȡ
 *  (4) ܥեȥѤˤľŪޤϴŪ뤤ʤ»
 *      ⡤嵭ԤTOPPERSץȤդ뤳ȡ
 *      ޤܥեȥΥ桼ޤϥɥ桼Τʤ
 *      ͳ˴Ťᤫ⡤嵭ԤTOPPERSץȤ
 *      դ뤳ȡ
 * 
 *  ܥեȥϡ̵ݾڤ󶡤ƤΤǤ롥嵭Ԥ
 *  TOPPERSץȤϡܥեȥ˴ؤơλŪ
 *  ФŬޤơʤݾڤԤʤޤܥեȥ
 *  ѤˤľŪޤϴŪʤ»˴ؤƤ⡤
 *  Ǥʤ
 * 
 */
/*!
 *  \file   toppers/diagnostics.hpp
 *  \brief  ǽ˴ؤ
 */
#ifndef TOPPERS_DIAGNOSTICS_HPP_
#define TOPPERS_DIAGNOSTICS_HPP_

#include <stdexcept>
#include "toppers/debug.hpp"
#include "toppers/gettext.hpp"
#include <boost/format.hpp>

namespace toppers
{

  struct text_line;

  class diagnostics_error : public std::runtime_error
  {
  public:
    diagnostics_error( std::string const& msg ) : std::runtime_error( msg ) {}
  };

  class normal_exit {};

  int get_error_count();
  int increment_error_count();
  void set_program_name( char const* name );
  std::string const& get_program_name();
  int set_error_abort_threshold( int thresh );
  void warning( char const* msg );
  void warning( text_line const& line, char const* msg );
  void error( char const* msg );
  void error( text_line const& line, char const* msg );
  void fatal( char const* msg );
  void fatal( text_line const& line, char const* msg );
  void set_error_location( char const* msg );

  template < typename T1 >
    inline void warning( char const* str, T1 const& arg1 )
  {
    warning( ( boost::format( str ) % arg1 ).str().c_str() );
  }

  template < typename T1, typename T2 >
    inline void warning( char const* str, T1 const& arg1, T2 const& arg2 )
  {
    warning( ( boost::format( str ) % arg1 % arg2 ).str().c_str() );
  }

  template < typename T1, typename T2, typename T3 >
    inline void warning( char const* str, T1 const& arg1, T2 const& arg2, T3 const& arg3 )
  {
    warning( ( boost::format( str ) % arg1 % arg2 % arg3 ).str().c_str() );
  }

  template < typename T1, typename T2, typename T3, typename T4 >
    inline void warning( char const* str, T1 const& arg1, T2 const& arg2, T3 const& arg3, T4 const& arg4 )
  {
    warning( ( boost::format( str ) % arg1 % arg2 % arg3 % arg4 ).str().c_str() );
  }

  template < typename T1 >
    inline void warning( text_line const& line, char const* str, T1 const& arg1 )
  {
    warning( line, ( boost::format( str ) % arg1 ).str().c_str() );
  }

  template < typename T1, typename T2 >
    inline void warning( text_line const& line, char const* str, T1 const& arg1, T2 const& arg2 )
  {
    warning( line, ( boost::format( str ) % arg1 % arg2 ).str().c_str() );
  }

  template < typename T1, typename T2, typename T3 >
    inline void warning( text_line const& line, char const* str, T1 const& arg1, T2 const& arg2, T3 const& arg3 )
  {
    warning( line, ( boost::format( str ) % arg1 % arg2 % arg3 ).str().c_str() );
  }

  template < typename T1, typename T2, typename T3, typename T4 >
    inline void warning( text_line const& line, char const* str, T1 const& arg1, T2 const& arg2, T3 const& arg3, T4 const& arg4 )
  {
    warning( line, ( boost::format( str ) % arg1 % arg2 % arg3 % arg4 ).str().c_str() );
  }

  template < typename T1 >
    inline void error( char const* str, T1 const& arg1 )
  {
    error( ( boost::format( str ) % arg1 ).str().c_str() );
  }

  template < typename T1, typename T2 >
    inline void error( char const* str, T1 const& arg1, T2 const& arg2 )
  {
    error( ( boost::format( str ) % arg1 % arg2 ).str().c_str() );
  }

  template < typename T1, typename T2, typename T3 >
    inline void error( char const* str, T1 const& arg1, T2 const& arg2, T3 const& arg3 )
  {
    error( ( boost::format( str ) % arg1 % arg2 % arg3 ).str().c_str() );
  }

  template < typename T1, typename T2, typename T3, typename T4 >
    inline void error( char const* str, T1 const& arg1, T2 const& arg2, T3 const& arg3, T4 const& arg4 )
  {
    error( ( boost::format( str ) % arg1 % arg2 % arg3 % arg4 ).str().c_str() );
  }

  template < typename T1, typename T2, typename T3, typename T4, typename T5 >
    inline void error( char const* str, T1 const& arg1, T2 const& arg2, T3 const& arg3, T4 const& arg4 , T5 const& arg5 )
  {
    error( ( boost::format( str ) % arg1 % arg2 % arg3 % arg4 % arg5 ).str().c_str() );
  }
  
  template < typename T1 >
    inline void error( text_line const& line, char const* str, T1 const& arg1 )
  {
    error( line, ( boost::format( str ) % arg1 ).str().c_str() );
  }

  template < typename T1, typename T2 >
    inline void error( text_line const& line, char const* str, T1 const& arg1, T2 const& arg2 )
  {
    error( line, ( boost::format( str ) % arg1 % arg2 ).str().c_str() );
  }

  template < typename T1, typename T2, typename T3 >
    inline void error( text_line const& line, char const* str, T1 const& arg1, T2 const& arg2, T3 const& arg3 )
  {
    error( line, ( boost::format( str ) % arg1 % arg2 % arg3 ).str().c_str() );
  }

  template < typename T1, typename T2, typename T3, typename T4 >
    inline void error( text_line const& line, char const* str, T1 const& arg1, T2 const& arg2, T3 const& arg3, T4 const& arg4 )
  {
    error( line, ( boost::format( str ) % arg1 % arg2 % arg3 % arg4 ).str().c_str() );
  }

  template < typename T1 >
    inline void fatal( char const* str, T1 const& arg1 )
  {
    fatal( ( boost::format( str ) % arg1 ).str().c_str() );
  }

  template < typename T1, typename T2 >
    inline void fatal( char const* str, T1 const& arg1, T2 const& arg2 )
  {
    fatal( ( boost::format( str ) % arg1 % arg2 ).str().c_str() );
  }

  template < typename T1, typename T2, typename T3 >
    inline void fatal( char const* str, T1 const& arg1, T2 const& arg2, T3 const& arg3 )
  {
    fatal( ( boost::format( str ) % arg1 % arg2 % arg3 ).str().c_str() );
  }

  template < typename T1, typename T2, typename T3, typename T4 >
    inline void fatal( char const* str, T1 const& arg1, T2 const& arg2, T3 const& arg3, T4 const& arg4 )
  {
    fatal( ( boost::format( str ) % arg1 % arg2 % arg3 % arg4 ).str().c_str() );
  }

  template < typename T1 >
    inline void fatal( text_line const& line, char const* str, T1 const& arg1 )
  {
    fatal( line, ( boost::format( str ) % arg1 ).str().c_str() );
  }

  template < typename T1, typename T2 >
    inline void fatal( text_line const& line, char const* str, T1 const& arg1, T2 const& arg2 )
  {
    fatal( line, ( boost::format( str ) % arg1 % arg2 ).str().c_str() );
  }

  template < typename T1, typename T2, typename T3 >
    inline void fatal( text_line const& line, char const* str, T1 const& arg1, T2 const& arg2, T3 const& arg3 )
  {
    fatal( line, ( boost::format( str ) % arg1 % arg2 % arg3 ).str().c_str() );
  }

  template < typename T1, typename T2, typename T3, typename T4 >
    inline void fatal( text_line const& line, char const* str, T1 const& arg1, T2 const& arg2, T3 const& arg3, T4 const& arg4 )
  {
    fatal( line, ( boost::format( str ) % arg1 % arg2 % arg3 % arg4 ).str().c_str() );
  }

  inline void exit()
  {
    throw normal_exit();
  }

#undef  _
#define _( str )  ::toppers::gettext( str ).c_str()

}

#endif  // ! TOPPERS_DIAGNOSTICS_HPP_

